@@ -1,5 +1,6 @@
 # mypy: ignore-errors
 import os
+from functools import lru_cache
 from typing import Optional, Union, Callable, List, Dict
 
 import numpy as np
@@ -8,7 +9,8 @@ import numpy.typing as npt
 from code_loader.contract.datasetclasses import CustomCallableInterfaceMultiArgs, \
     CustomMultipleReturnCallableInterfaceMultiArgs, ConfusionMatrixCallableInterfaceMultiArgs, CustomCallableInterface, \
     VisualizerCallableInterface, MetadataSectionCallableInterface, PreprocessResponse, SectionCallableInterface, \
-    ConfusionMatrixElement, SamplePreprocessResponse, PredictionTypeHandler, InstanceCallableInterface, ElementInstance
+    ConfusionMatrixElement, SamplePreprocessResponse, PredictionTypeHandler, InstanceCallableInterface, ElementInstance, \
+    InstanceLengthCallableInterface
 from code_loader.contract.enums import MetricDirection, LeapDataType, DatasetMetadataType
 from code_loader import leap_binder
 from code_loader.contract.mapping import NodeMapping, NodeMappingType, NodeConnection
@@ -20,6 +22,7 @@ import inspect
 import functools
 
 _called_from_inside_tl_decorator = 0
+_called_from_inside_tl_integration_test_decorator = False
 
 
 def _add_mapping_connection(user_unique_name, connection_destinations, arg_names, name, node_mapping_type):
@@ -45,17 +48,22 @@ def integration_test():
         leap_binder.integration_test_func = integration_test_function
 
         def inner(*args, **kwargs):
-            ret = integration_test_function(*args, **kwargs)
-
+            global _called_from_inside_tl_integration_test_decorator
             try:
-                os.environ[mapping_runtime_mode_env_var_mame] = 'True'
-                integration_test_function(None, None)
-            except Exception as e:
-                print(f'Error during integration test: Make sure to disable any non tensorleap decorators '
-                      f'functions before pushing a new TL version')
+                _called_from_inside_tl_integration_test_decorator = True
+                ret = integration_test_function(*args, **kwargs)
+
+                try:
+                    os.environ[mapping_runtime_mode_env_var_mame] = 'True'
+                    integration_test_function(None, None)
+                except Exception as e:
+                    print(f'Error during integration test: Make sure to disable any non tensorleap decorators '
+                          f'functions before pushing a new TL version')
+                finally:
+                    if mapping_runtime_mode_env_var_mame in os.environ:
+                        del os.environ[mapping_runtime_mode_env_var_mame]
             finally:
-                if mapping_runtime_mode_env_var_mame in os.environ:
-                    del os.environ[mapping_runtime_mode_env_var_mame]
+                _called_from_inside_tl_integration_test_decorator = False
 
         return inner
 
@@ -70,6 +78,7 @@ def tensorleap_load_model(prediction_types: Optional[List[PredictionTypeHandler]
         class TempMapping:
             pass
 
+        @lru_cache()
         def inner():
             class ModelPlaceholder:
                 def __init__(self):
@@ -84,6 +93,9 @@ def tensorleap_load_model(prediction_types: Optional[List[PredictionTypeHandler]
                 def run(self, output_names, input_dict):
                     return self.model.run(output_names, input_dict)
 
+                def get_inputs(self):
+                    return self.model.get_inputs()
+
             return ModelPlaceholder()
 
         def mapping_inner():
@@ -115,11 +127,31 @@ def tensorleap_load_model(prediction_types: Optional[List[PredictionTypeHandler]
                     assert output_names is None
                     assert isinstance(input_dict, dict), \
                         f'Expected input_dict to be a dict, got {type(input_dict)} instead.'
-                    for i, elem in enumerate(input_dict.values()):
-                        elem.node_mapping.type = NodeMappingType[f'Input{str(i)}']
+                    for i, (input_key, elem) in enumerate(input_dict.items()):
+                        if isinstance(input_key, NodeMappingType):
+                            elem.node_mapping.type = input_key
+                        else:
+                            elem.node_mapping.type = NodeMappingType[f'Input{str(i)}']
 
                     return ModelOutputPlaceholder()
 
+                def get_inputs(self):
+                    class FollowIndex:
+                        def __init__(self, index):
+                            self.name =  NodeMappingType[f'Input{str(i)}']
+
+                    class FollowInputIndex:
+                        def __init__(self):
+                            pass
+
+                        def __getitem__(self, index):
+                            assert isinstance(index, int), \
+                                f'Expected key to be an int, got {type(index)} instead.'
+
+                            return FollowIndex(index)
+
+                    return FollowInputIndex()
+
             return ModelPlaceholder()
 
         def final_inner():
@@ -529,29 +561,24 @@ def tensorleap_preprocess():
 
 
 def tensorleap_element_instance_preprocess(
-        instance_mask_encoder: Callable[[str, PreprocessResponse], List[ElementInstance]]):
+        instance_length_encoder: InstanceLengthCallableInterface):
     def decorating_function(user_function: Callable[[], List[PreprocessResponse]]):
         def user_function_instance() -> List[PreprocessResponse]:
             result = user_function()
             for preprocess_response in result:
                 sample_ids_to_instance_mappings = {}
                 instance_to_sample_ids_mappings = {}
-                instance_ids_to_names = {}
                 all_sample_ids = preprocess_response.sample_ids.copy()
                 for sample_id in preprocess_response.sample_ids:
-                    instances_masks = instance_mask_encoder(sample_id, preprocess_response)
-                    instances_ids = [f'{sample_id}_{instance_id}' for instance_id in range(len(instances_masks))]
+                    instances_length = instance_length_encoder(sample_id, preprocess_response)
+                    instances_ids = [f'{sample_id}_{instance_id}' for instance_id in range(instances_length)]
                     sample_ids_to_instance_mappings[sample_id] = instances_ids
                     instance_to_sample_ids_mappings[sample_id] = sample_id
-                    instance_names = [instance.name for instance in instances_masks]
-                    instance_ids_to_names[sample_id] = 'none'
-                    for instance_id, instance_name in zip(instances_ids, instance_names):
+                    for instance_id in instances_ids:
                         instance_to_sample_ids_mappings[instance_id] = sample_id
-                        instance_ids_to_names[instance_id] = instance_name
                     all_sample_ids.extend(instances_ids)
                 preprocess_response.sample_ids_to_instance_mappings = sample_ids_to_instance_mappings
                 preprocess_response.instance_to_sample_ids_mappings = instance_to_sample_ids_mappings
-                preprocess_response.instance_ids_to_names = instance_ids_to_names
                 preprocess_response.sample_ids = all_sample_ids
             return result
 
@@ -620,7 +647,7 @@ def tensorleap_unlabeled_preprocess():
 
 def tensorleap_instances_masks_encoder(name: str):
     def decorating_function(user_function: InstanceCallableInterface):
-        def _validate_input_args(sample_id: str, preprocess_response: PreprocessResponse):
+        def _validate_input_args(sample_id: str, preprocess_response: PreprocessResponse, instance_id: int):
             assert isinstance(sample_id, str), \
                 (f'tensorleap_instances_masks_encoder validation failed: '
                  f'Argument sample_id should be str. Got {type(sample_id)}.')
@@ -631,11 +658,61 @@ def tensorleap_instances_masks_encoder(name: str):
                 (f'tensorleap_instances_masks_encoder validation failed: '
                  f'Argument sample_id should be as the same type as defined in the preprocess response '
                  f'{preprocess_response.sample_id_type}. Got {type(sample_id)}.')
+            assert isinstance(instance_id, int), \
+                (f'tensorleap_instances_masks_encoder validation failed: '
+                 f'Argument instance_id should be int. Got {type(instance_id)}.')
 
         def _validate_result(result):
-            assert isinstance(result, list), \
+            assert isinstance(result, ElementInstance) or (result is None), \
                 (f'tensorleap_instances_masks_encoder validation failed: '
-                 f'Unsupported return type. Should be a numpy array. Got {type(result)}.')
+                 f'Unsupported return type. Should be a ElementInstance or None. Got {type(result)}.')
+
+        def inner_without_validate(sample_id, preprocess_response, instance_id):
+            global _called_from_inside_tl_decorator
+            _called_from_inside_tl_decorator += 1
+
+            try:
+                result = user_function(sample_id, preprocess_response, instance_id)
+            finally:
+                _called_from_inside_tl_decorator -= 1
+
+            return result
+
+        leap_binder.set_instance_masks(inner_without_validate, name)
+
+        def inner(sample_id, preprocess_response, instance_id):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return None
+
+            _validate_input_args(sample_id, preprocess_response, instance_id)
+
+            result = inner_without_validate(sample_id, preprocess_response, instance_id)
+
+            _validate_result(result)
+            return result
+
+        return inner
+
+    return decorating_function
+
+def tensorleap_instances_length_encoder(name: str):
+    def decorating_function(user_function: InstanceLengthCallableInterface):
+        def _validate_input_args(sample_id: str, preprocess_response: PreprocessResponse):
+            assert isinstance(sample_id, str), \
+                (f'tensorleap_instances_length_encoder validation failed: '
+                 f'Argument sample_id should be str. Got {type(sample_id)}.')
+            assert isinstance(preprocess_response, PreprocessResponse), \
+                (f'tensorleap_instances_length_encoder validation failed: '
+                 f'Argument preprocess_response should be a PreprocessResponse. Got {type(preprocess_response)}.')
+            assert type(sample_id) == preprocess_response.sample_id_type, \
+                (f'tensorleap_instances_length_encoder validation failed: '
+                 f'Argument sample_id should be as the same type as defined in the preprocess response '
+                 f'{preprocess_response.sample_id_type}. Got {type(sample_id)}.')
+
+        def _validate_result(result):
+            assert isinstance(result, int), \
+                (f'tensorleap_instances_length_encoder validation failed: '
+                 f'Unsupported return type. Should be a int. Got {type(result)}.')
 
         def inner_without_validate(sample_id, preprocess_response):
             global _called_from_inside_tl_decorator
@@ -648,8 +725,6 @@ def tensorleap_instances_masks_encoder(name: str):
 
             return result
 
-        leap_binder.set_instance_masks(inner_without_validate, name)
-
         def inner(sample_id, preprocess_response):
             if os.environ.get(mapping_runtime_mode_env_var_mame):
                 return None
@@ -665,7 +740,6 @@ def tensorleap_instances_masks_encoder(name: str):
 
     return decorating_function
 
-
 def tensorleap_input_encoder(name: str, channel_dim=-1, model_input_index=None):
     def decorating_function(user_function: SectionCallableInterface):
         for input_handler in leap_binder.setup_container.inputs:
@@ -719,7 +793,7 @@ def tensorleap_input_encoder(name: str, channel_dim=-1, model_input_index=None):
 
             _validate_result(result)
 
-            if _called_from_inside_tl_decorator == 0:
+            if _called_from_inside_tl_decorator == 0 and _called_from_inside_tl_integration_test_decorator:
                 result = np.expand_dims(result, axis=0)
 
             return result
@@ -803,7 +877,7 @@ def tensorleap_gt_encoder(name: str):
 
             _validate_result(result)
 
-            if _called_from_inside_tl_decorator == 0:
+            if _called_from_inside_tl_decorator == 0 and _called_from_inside_tl_integration_test_decorator:
                 result = np.expand_dims(result, axis=0)
 
             return result
